<?php

/**
 * @package Auto-Install Free SSL
 * This package is a WordPress Plugin. It issues and installs free SSL certificates in cPanel shared hosting with complete automation.
 *
 * @author Free SSL Dot Tech <support@freessl.tech>
 * @copyright  Copyright (C) 2019-2020, Anindya Sundar Mandal
 * @license    http://www.gnu.org/licenses/gpl-3.0.html GNU General Public License, version 3
 * @link       https://freessl.tech
 * @since      Class available since Release 1.0.0
 *
 *
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 */
namespace AutoInstallFreeSSL\FreeSSLAuto;

use  DateTime ;
use  AutoInstallFreeSSL\FreeSSLAuto\Acme\Factory ;
use  AutoInstallFreeSSL\FreeSSLAuto\Admin\HomeOptions ;
use  Exception ;
//Common actions, even if the control panel is not cPanel
class Controller
{
    /**
     * Initiates the Controller class.
     */
    public function __construct()
    {
        if ( !defined( 'ABSPATH' ) ) {
            die( __( "Access Denied", 'auto-install-free-ssl' ) );
        }
        $this->logger = new Logger();
    }
    
    /**
     *
     *
     * Make array of the domains pointing to the same document root of a domain.
     *
     * @param array $single_domain
     * @param array $domains_to_exclude
     *
     * @return array
     */
    public function domainsArray( $single_domain, $domains_to_exclude )
    {
        $domains_array = [];
        if ( !\in_array( $single_domain['domain'], $domains_to_exclude, true ) || $this->sslRequiredForFirstTime( $single_domain['domain'] ) ) {
            $domains_array[] = $single_domain['domain'];
        }
        
        if ( \strlen( $single_domain['serveralias'] ) > 1 ) {
            $domains = explode( ' ', $single_domain['serveralias'] );
            foreach ( $domains as $domain ) {
                //Exclude $domains_to_exclude
                if ( !\in_array( $domain, $domains_to_exclude, true ) || $this->sslRequiredForFirstTime( $domain ) ) {
                    if ( $single_domain['domain'] !== $domain ) {
                        $domains_array[] = $domain;
                    }
                }
            }
        }
        
        //remove offline domains
        //e.g., in Godaddy hosting www version of any sub domain is offline by default
        $domains_online = [];
        foreach ( $domains_array as $key => $domain ) {
            
            if ( false === strpos( $domain, '*.' ) ) {
                //check if domain is online, only for non-wildcard domains
                $socket = @fsockopen(
                    $domain,
                    80,
                    $errno,
                    $errstr,
                    30
                );
                
                if ( $socket ) {
                    //Domain is online
                    $domains_online[] = $domain;
                } else {
                    //domain offline
                    $this->logger->log( $domain . ' is offline. Skipping it.' );
                }
            
            } else {
                //Domain is wildcard and will be validated by DNS-01 challange, so, online check not required
                $domains_online[] = $domain;
            }
        
        }
        return $domains_online;
    }
    
    /**
     * Checks whether SSL certificate MUST required for the given domain
     * @param $domain
     *
     * @return bool
     */
    public function sslRequiredForFirstTime( $domain )
    {
        //For free version always return false
        return false;
    }
    
    /**
     *
     *
     * Get the wildcard base domain.
     *
     * @param string $domain_as_is
     *
     * @return string
     */
    public function getWildcardBase( $domain_as_is )
    {
        $registeredDomain = getRegisteredDomain( $domain_as_is );
        if ( null === $registeredDomain ) {
            return false;
        }
        //compute wildcard domain
        
        if ( \strlen( $domain_as_is ) > \strlen( $registeredDomain ) ) {
            //may be it's a subdomain
            $part = str_replace( $registeredDomain, '', $domain_as_is );
            $domain_elements_array = explode( '.', $part );
            if ( 2 === \count( $domain_elements_array ) && 'www' === $domain_elements_array[0] ) {
                //www.domain.com is not considered as wildcard domain
                return $domain_as_is;
            }
            //get the position of first . and ONLY replace the part left to it with *
            $pos = strpos( $domain_as_is, '.' );
            return '*' . substr( $domain_as_is, $pos );
        } elseif ( \strlen( $domain_as_is ) === \strlen( $registeredDomain ) ) {
            return $domain_as_is;
        }
    
    }

}